// $Id: vector.cc 69 2006-02-11 00:44:22Z jari $

#include <c++_tools/gslapi/vector.h>
#include <c++_tools/utility/stl_utility.h>

#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <utility>


namespace theplu {
namespace gslapi {



	vector::vector(void)
		: v_(NULL), view_(NULL)
	{
	}



	vector::vector(const size_t n,const double init_value)
		: view_(NULL)
	{
		v_ = gsl_vector_alloc(n);
		set_all(init_value);
	}



	vector::vector(const vector& other)
		: view_(NULL)
	{
		v_ = other.TEMP_gsl_vector_copy();
	}



	vector::vector(vector& v, size_t offset, size_t n, size_t stride)
	{
		view_ = new gsl_vector_view(gsl_vector_subvector_with_stride(v.v_,offset,
																																 stride,n));
		v_ = &(view_->vector);
	}



	vector::vector(gsl_vector* vec)
		: v_(vec), view_(NULL)
	{
	}



	vector::vector(std::istream& is) throw (utility::IO_error,std::exception)
		: view_(NULL)
	{
		// read the data file and store in stl vectors (dynamically
		// expandable)
		std::vector<std::vector<double> > data_matrix;
		u_int nof_columns=0;
		u_int nof_rows=0;
		std::vector<double> v;
		for (nof_rows = 0; utility::read_to_double(is, v); nof_rows++) {
			// Ignoring empty lines
			if (!v.size()) {
				nof_rows--;
				continue;
			}
			if (!nof_columns)
				nof_columns=v.size();
			else if (nof_rows && (nof_columns>1)) {
				std::ostringstream s;
				s << "vector::vector(std::istream&) data file error:\n"
					<< "    File has inconsistent number of rows (" << nof_rows+1
					<< ") and columns (" << nof_columns
					<< ").\n    Expected a row or a column vector.";
				std::string m=s.str();
				throw utility::IO_error(m);
			}
			else if(v.size()!=nof_columns) {
				std::ostringstream s;
				s << "vector::vector(std::istream&) data file error:\n"
					<< "    Line " << (nof_rows+1) << " has " << v.size()
					<< " columns; expected " << nof_columns << " column.";
				std::string m=s.str();
				throw utility::IO_error(m);
			}
			data_matrix.push_back(v);
		}

		// manipulate the state of the stream to be good
		is.clear(std::ios::goodbit);
		// convert the data to a gsl vector
		v_ = gsl_vector_alloc(nof_rows*nof_columns);
		size_t n=0;
		for (size_t i=0; i<nof_rows; i++)
			for (size_t j=0; j<nof_columns; j++) 
				gsl_vector_set( v_, n++, data_matrix[i][j] );
	}



	vector::~vector(void)
	{
		if (view_)
			delete view_;
		else if (v_)
			gsl_vector_free(v_);
		v_=NULL;
	}



	gsl_vector* vector::TEMP_gsl_vector_copy(void) const
	{
		gsl_vector* vec = gsl_vector_alloc(size());
		gsl_vector_memcpy(vec,v_);
		return vec;
	}



	std::pair<double,double> vector::minmax(void) const
	{
		double min, max;
		gsl_vector_minmax(v_,&min,&max);
		return std::pair<double,double>(min,max);
	}



	std::pair<size_t,size_t> vector::minmax_index(void) const
	{
		size_t min_index, max_index;
		gsl_vector_minmax_index(v_,&min_index,&max_index);
		return std::pair<size_t,size_t>(min_index,max_index);
	}



	double vector::sum(void) const
	{
		double sum = 0;
		for (size_t i=0; i<size(); i++)
			sum += gsl_vector_get( v_, i );
		return( sum );
	}  



	double vector::operator*( const vector &other ) const
	{
		// Jari, check for gsl_support
		double res = 0.0;;
		for ( size_t i = 0; i < size(); ++i ) 
			res += other[i] * (*this)[i];
		return res;
	}



	vector vector::operator+(const vector &other) const
	{
		vector res(*this);
		gsl_vector_add(res.v_,other.v_);
		return res;
	}



	vector vector::operator-( const vector &other ) const
	{
		vector res( *this );
		gsl_vector_sub(res.v_,other.v_);
		return res;
	}



	bool vector::operator==(const vector& a) const
	{
		if (size()!=a.size())
			return false;
		for (size_t i=0; i<size(); ++i)
			if (gsl_vector_get(v_,i)!=a(i))
				return false;
		return true;
	}



  const vector& vector::operator=( const vector& other )
	{
		if( this != &other ) {
			if ( v_ )
				gsl_vector_free( v_ );
			v_ = other.TEMP_gsl_vector_copy();
    }
		return *this;
	} 



	std::ostream& operator<<(std::ostream& s, const vector& a)
	{
		s.setf(std::ios::fixed);
		s.precision(12);
		for (size_t j = 0; j < a.size(); ++j) {
			s << a[j];
			if ( (j+1)<a.size() )
				s << " ";
		}

		return s;
	}


}} // of namespace gslapi and namespace thep
